@@ -6,14 +6,16 @@ from decimal import Decimal
 from simpleeval import simple_eval
 
 from trytond.i18n import gettext
-from trytond.pool import PoolMeta, Pool
 from trytond.model import (
-    ModelSQL, ModelView, MatchMixin, Workflow, DeactivableMixin, fields)
-from trytond.pyson import Eval, If, Bool
-from trytond.transaction import Transaction
-from trytond.tools import decistmt
-
+    DeactivableMixin, MatchMixin, ModelSQL, ModelView, Workflow, fields)
+from trytond.modules.currency.fields import Monetary
 from trytond.modules.product import price_digits, round_price
+from trytond.modules.product_price_list import Null
+from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Bool, Eval, If
+from trytond.tools import decistmt
+from trytond.transaction import Transaction
+
 from .exceptions import FormulaError
 
 
@@ -76,7 +78,8 @@ class Line(metaclass=PoolMeta):
         ondelete='RESTRICT',
         domain=[
             ('company', '=', Eval('_parent_sale', {}).get('company', -1)),
-            ])
+            ],
+        depends={'sale'})
 
 
 class Promotion(
@@ -86,7 +89,7 @@ class Promotion(
 
     name = fields.Char('Name', translate=True, required=True)
     company = fields.Many2One(
-        'company.company', "Company", required=True, select=True,
+        'company.company', "Company", required=True,
         states={
             'readonly': Eval('id', 0) > 0,
             })
@@ -96,63 +99,50 @@ class Promotion(
                     datetime.date.max,
                     Eval('end_date', datetime.date.max))),
             ('start_date', '=', None),
-            ],
-        depends=['end_date'])
+            ])
     end_date = fields.Date('End Date',
         domain=['OR',
             ('end_date', '>=', If(~Eval('start_date', None),
                     datetime.date.min,
                     Eval('start_date', datetime.date.min))),
             ('end_date', '=', None),
-            ],
-        depends=['start_date'])
+            ])
     price_list = fields.Many2One('product.price_list', 'Price List',
         ondelete='CASCADE',
         domain=[
             ('company', '=', Eval('company', -1)),
-            ],
-        depends=['company'])
+            ])
 
-    amount = fields.Numeric(
-        "Amount", digits=(16, Eval('currency_digits', 2)),
-        depends=['currency_digits'])
+    amount = Monetary("Amount", currency='currency', digits='currency')
     currency = fields.Many2One(
         'currency.currency', "Currency",
         states={
             'required': Bool(Eval('amount', 0)),
-            },
-        depends=['amount'])
-    currency_digits = fields.Function(
-        fields.Integer("Currency Digits"), 'on_change_with_currency_digits')
+            })
     untaxed_amount = fields.Boolean(
         "Untaxed Amount",
         states={
             'invisible': ~Eval('amount'),
-            },
-        depends=['amount'])
+            })
 
-    quantity = fields.Float('Quantity', digits=(16, Eval('unit_digits', 2)),
-        depends=['unit_digits'])
+    quantity = fields.Float('Quantity', digits='unit')
     unit = fields.Many2One('product.uom', 'Unit',
         states={
             'required': Bool(Eval('quantity', 0)),
-            },
-        depends=['quantity'])
-    unit_digits = fields.Function(fields.Integer('Unit Digits'),
-        'on_change_with_unit_digits')
+            })
     products = fields.Many2Many(
         'sale.promotion-product.product', 'promotion', 'product', "Products",
         context={
             'company': Eval('company', -1),
             },
-        depends=['company'])
+        depends={'company'})
     categories = fields.Many2Many(
         'sale.promotion-product.category', 'promotion', 'category',
         "Categories",
         context={
             'company': Eval('company', -1),
             },
-        depends=['company'])
+        depends={'company'})
     formula = fields.Char('Formula', required=True,
         help=('Python expression that will be evaluated with:\n'
             '- unit_price: the original unit_price'))
@@ -165,40 +155,34 @@ class Promotion(
     def default_untaxed_amount(cls):
         return False
 
-    @fields.depends('unit')
-    def on_change_with_unit_digits(self, name=None):
-        if self.unit:
-            return self.unit.digits
-        return 2
-
-    @fields.depends('currency')
-    def on_change_with_currency_digits(self, name=None):
-        if self.currency:
-            return self.currency.digits
+    @classmethod
+    def validate_fields(cls, promotions, field_names):
+        super().validate_fields(promotions, field_names)
+        cls.check_formula(promotions, field_names)
 
     @classmethod
-    def validate(cls, promotions):
-        super().validate(promotions)
+    def check_formula(cls, promotions, field_names=None):
+        if field_names and 'formula' not in field_names:
+            return
         for promotion in promotions:
-            promotion.check_formula()
-
-    def check_formula(self):
-        context = self.get_context_formula(None)
-        try:
-            if not isinstance(self.get_unit_price(**context), Decimal):
-                raise ValueError('Not a Decimal')
-        except Exception as exception:
-            raise FormulaError(
-                gettext('sale_promotion.msg_invalid_formula',
-                    formula=self.formula,
-                    promotion=self.rec_name,
-                    exception=exception)) from exception
+            context = promotion.get_context_formula(None)
+            try:
+                unit_price = promotion.get_unit_price(**context)
+                if not isinstance(unit_price, Decimal):
+                    raise ValueError('Not a Decimal')
+            except Exception as exception:
+                raise FormulaError(
+                    gettext('sale_promotion.msg_invalid_formula',
+                        formula=promotion.formula,
+                        promotion=promotion.rec_name,
+                        exception=exception)) from exception
 
     @classmethod
     def _promotions_domain(cls, sale):
         pool = Pool()
         Date = pool.get('ir.date')
-        sale_date = sale.sale_date or Date.today()
+        with Transaction().set_context(company=sale.company.id):
+            sale_date = sale.sale_date or Date.today()
         return [
             ['OR',
                 ('start_date', '<=', sale_date),
@@ -301,12 +285,13 @@ class Promotion(
                 continue
             context = self.get_context_formula(line)
             new_price = self.get_unit_price(**context)
-            if new_price < 0:
-                new_price = Decimal(0)
-            if line.unit_price >= new_price:
-                line.unit_price = round_price(new_price)
-                line.promotion = self
-                applied = True
+            if new_price is not None:
+                if new_price < 0:
+                    new_price = Decimal(0)
+                if line.unit_price >= new_price:
+                    line.unit_price = round_price(new_price)
+                    line.promotion = self
+                    applied = True
         if applied:
             sale.lines = sale.lines  # Trigger the change
 
@@ -322,22 +307,26 @@ class Promotion(
             unit_price = Decimal(0)
         return {
             'names': {
-                'unit_price': unit_price,
+                'unit_price': unit_price if unit_price is not None else Null(),
                 },
             }
 
     def get_unit_price(self, **context):
         'Return unit price (as Decimal)'
         context.setdefault('functions', {})['Decimal'] = Decimal
-        return max(simple_eval(decistmt(self.formula), **context), Decimal(0))
+        unit_price = simple_eval(decistmt(self.formula), **context)
+        unit_price = max(unit_price, Decimal(0))
+        if isinstance(unit_price, Null):
+            unit_price = None
+        return unit_price
 
 
 class Promotion_Product(ModelSQL):
     'Sale Promotion - Product'
     __name__ = 'sale.promotion-product.product'
 
-    promotion = fields.Many2One('sale.promotion', 'Promotion',
-        required=True, ondelete='CASCADE', select=True)
+    promotion = fields.Many2One(
+        'sale.promotion', "Promotion", required=True, ondelete='CASCADE')
     product = fields.Many2One('product.product', 'Product',
         required=True, ondelete='CASCADE')
 
@@ -347,8 +336,7 @@ class Promotion_ProductCategory(ModelSQL):
     __name__ = 'sale.promotion-product.category'
 
     promotion = fields.Many2One(
-        'sale.promotion', "Promotion",
-        required=True, ondelete='CASCADE', select=True)
+        'sale.promotion', "Promotion", required=True, ondelete='CASCADE')
     category = fields.Many2One(
         'product.category', "Category",
         required=True, ondelete='CASCADE')
